#!/usr/bin/python

import os

import matplotlib
matplotlib.use('GTKAgg')
import matplotlib.pyplot as pyplot
from matplotlib.backends.backend_gtkagg import NavigationToolbar2GTKAgg as NavigationToolbar
from matplotlib.backends.backend_gtkagg import FigureCanvasGTKAgg as FigureCanvas

import numpy

from savitzkygolay import savitzky_golay


class DataSet:

	class FileFormatException(Exception):
		pass

	def __init__(self, filename, do_smooth, smooth_kernel, smooth_order):
		self.properties = {}
		self.name, _ = os.path.splitext(os.path.basename(filename))

		with open(filename, 'r') as f:
			if f.next().rstrip() != "Strain":
				raise self.__class__.FileFormatException()

			for line in f:
				line = line.strip()
				if line == 'DATA':
					break
				else:
					key,val = line.split(',')
					self.properties[key] = val

			self.step = float(self.properties['RESOLUTION[m]'])
			self.len = int(self.properties['DATA NUM'])
			self.y = numpy.fromiter(f, numpy.float, self.len)
			self.diff = numpy.fabs(numpy.diff(self.y))

			self.locate_end()
			self.smooth(enabled=do_smooth, kernel=smooth_kernel, order=smooth_order)

	def locate_end(self):
		# TODO: this heuristic is just a quick hack and doesn't work well
		threshold = numpy.max(self.diff) * 0.4
		self.end = numpy.argwhere(self.diff > threshold)[0] - 1

	def smooth(self, order, kernel, enabled):
		if enabled:
			if not hasattr(self, 'y_orig'):
				self.y_orig = self.y
			self.y = savitzky_golay(self.y_orig, kernel, order)
		else:
			if hasattr(self, 'y_orig'):
				self.y = self.y_orig


class AverageDataSet:

	def __init__(self, dss, name):
		# check that the data sets match
		self.len = dss[0].len
		self.step = dss[0].step
		for ds in dss:
			assert(self.len == ds.len and self.step == ds.step)

		self.y = numpy.average(numpy.array([ds.y for ds in dss]), axis=0)

		self.end = max([ds.end for ds in dss])
		self.name = name


class TemperatureCompensatedDataSet:

	def __init__(self, strain, temperature, temperature_factor, name):
		assert(strain.len == temperature.len and strain.step == temperature.step)
		self.len = strain.len
		self.step = strain.step
		self.y = strain.y - temperature.y*temperature_factor
		assert(len(self.y) == self.len)

		self.end = min(strain.end, temperature.end)
		self.name = name


class StandaloneGraph:

	def __init__(self, ds, extradss=[]):
		fig = pyplot.figure()
		axes = fig.add_subplot(1,1,1)
		axes.grid(True)
		axes.axvline(x=0, color='black')
		axes.axhline(y=0, color='black')
		axes.set_xlabel("Distance (m)")
		axes.set_ylabel("Strain (%)")

		axes.plot([i*ds.step for i in range(ds.len)], ds.y[:ds.len], label=ds.name)

		for xds in extradss:
			axes.plot([i*xds.step for i in range(xds.len)], xds.y[:xds.len], label=xds.name)

		xmax = ds.end

		ymax = ds.y[:xmax].max()
		ymin = ds.y[:xmax].min()
		axes.axvline(xmax, color='r')
		axes.set_xlim(0, xmax)
		axes.set_ylim(ymin - (ymax-ymin)*0.1, ymax + (ymax-ymin)*0.1)

		leg = axes.legend(labelspacing=0)
		for t in leg.get_texts():
			t.set_fontsize('small')

	def show(self):
		pyplot.show()
