#!/usr/bin/python

import os
import time
import gtk
from functools import partial

from datahandling import *


def relfile(path):
	return os.path.join(os.path.dirname(__file__), path)

class StrainAnalyserMainUI:

	def __init__(self):
		self.builder = gtk.Builder()
		self.builder.add_from_file(relfile("mainui.xml"))
		self.builder.connect_signals(self)

		self.project = StrainAnalyserProject()

		# initialise strain dataset list model-view-controller stuff
		# (insert "GTK, you're doing this wrong" soundbite)
		strainlist = self.builder.get_object("strainlist")
		straincols = (
				("Strain data file", 0),
				("Loss data file", 1),
				("Resolution (m)", 2)
			)
		for (title, columnId) in straincols:
			column = gtk.TreeViewColumn(title, gtk.CellRendererText(), text=columnId)
			column.set_resizable(True)
			column.set_sort_column_id(columnId)
			column.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
			column.set_expand(True)
			strainlist.append_column(column)
		strainlist.set_model(self.project.straindatasets)

		# initialise temperature dataset list model-view-controller stuff
		# (play "GTK, you're doing this wrong" soundbite on loop)
		temperaturelist = self.builder.get_object("temperaturelist")
		temperaturecols = (
				("Temperature data file", 0),
				("Resolution (m)", 2)
			)
		for (title, columnId) in temperaturecols:
			column = gtk.TreeViewColumn(title, gtk.CellRendererText(), text=columnId)
			column.set_resizable(True)
			column.set_sort_column_id(columnId)
			column.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
			column.set_expand(True)
			temperaturelist.append_column(column)
		temperaturelist.set_model(self.project.temperaturedatasets)

		# set some default values which seem to be ignored when set from Glade...
		self.builder.get_object("temperaturefactor").set_value(23.5/28)
		self.builder.get_object("noisefilt_order").set_value(5)
		self.builder.get_object("noisefilt_kernel").set_value(51)


	def show(self):
		self.builder.get_object("projectwindow").show()

	def quit(self):
		# TODO: project save prompt
		gtk.main_quit()


	# Signal handlers

	def on_projectwindow_destroy(self, widget):
		self.quit()

	def on_menu_quit_activate(self, widget):
		self.quit()

	def on_menu_about_activate(self, widget):
		aboutdialog = self.builder.get_object("aboutdialog")
		aboutdialog.run()  # synchronous: waits for box to be closed
		aboutdialog.hide()

	def on_add_strain_file_clicked(self, widget):
		chooserbuilder = gtk.Builder()
		chooserbuilder.add_from_file(relfile("straindatachooser.xml"))
		dialog = chooserbuilder.get_object("straindatachooser")
		response = dialog.run()
		if response == 0:
			self.project.add_strain_file(chooserbuilder.get_object("strainfile").get_filename(), chooserbuilder.get_object("lossfile").get_filename())
		dialog.destroy()

	def on_remove_strain_file_clicked(self, widget):
		strainlist = self.builder.get_object("strainlist")
		index = strainlist.get_cursor()[0][0]
		self.project.remove_strain_file(index)

	def on_add_temperature_file_clicked(self, widget):
		chooserbuilder = gtk.Builder()
		chooserbuilder.add_from_file(relfile("temperaturedatachooser.xml"))
		dialog = chooserbuilder.get_object("temperaturedatachooser")
		response = dialog.run()
		if response == 0:
			self.project.add_temperature_file(chooserbuilder.get_object("temperaturefile").get_filename())
		dialog.destroy()

	def on_remove_temperature_file_clicked(self, widget):
		temperaturelist = self.builder.get_object("temperaturelist")
		index = temperaturelist.get_cursor()[0][0]
		self.project.remove_temperature_file(index)

	def on_noisefilt_toggled(self, widget):
		state = widget.get_active()
		for target in ("noisefilt_order_slider", "noisefilt_kernel_slider", "noisefilt_order_label", "noisefilt_kernel_label"):
			self.builder.get_object(target).set_sensitive(state)

	def on_go_clicked(self, widget):
		waitboxbuilder = gtk.Builder()
		waitboxbuilder.add_from_file(relfile("waitbox.xml"))
		waitbox = waitboxbuilder.get_object("waitbox")
		statuslabel = waitboxbuilder.get_object("waitboxlabel")
		waitbox.show()
		while gtk.events_pending():
			gtk.main_iteration(False)

		def setstatus(label):
			statuslabel.set_label(label)
			time.sleep(0.1)
			while gtk.events_pending():
				gtk.main_iteration(False)

		self.project.run(
				status_cb=setstatus,
				finished_cb=waitbox.destroy,
				sg_enabled=self.builder.get_object("noisefilt").get_active(),
				sg_kernel=self.builder.get_object("noisefilt_kernel").get_value(),
				sg_order=self.builder.get_object("noisefilt_order").get_value(),
				temperature_factor=self.builder.get_object("temperaturefactor").get_value()
			)


class StrainAnalyserProject:

	def __init__(self):
		self.straindatasets = gtk.ListStore(str, str, float)
		self.temperaturedatasets = gtk.ListStore(str, float)

	def add_strain_file(self, strainfile, lossfile):
		assert(strainfile != None)
		# TODO: retrieve resolution
		self.straindatasets.append([strainfile, lossfile, 0])

	def remove_strain_file(self, index):
		self.straindatasets.remove(self.straindatasets.iter_nth_child(None, index))

	def add_temperature_file(self, temperaturefile):
		assert(temperaturefile != None)
		# TODO: retrieve resolution
		self.temperaturedatasets.append([temperaturefile, 0])

	def remove_temperature_file(self, index):
		self.temperaturedatasets.remove(self.temperaturedatasets.iter_nth_child(None, index))

	def run(self, sg_enabled=False, sg_kernel=None, sg_order=None, temperature_factor=1, status_cb=lambda msg: True, finished_cb=lambda: True):
		status_cb("Loading strain data")
		self.strain_dss = []
		# TODO: something with the loss data...
		for strainfile, lossfile, resolution in self.straindatasets:
			status_cb("Loading strain data: %s" % os.path.basename(strainfile))
			self.strain_dss += [DataSet(strainfile, do_smooth=sg_enabled, smooth_kernel=sg_kernel, smooth_order=sg_order)]

		status_cb("Loading temperature data")
		self.temperature_dss = []
		for temperaturefile, resolution in self.temperaturedatasets:
			status_cb("Loading temperature data: %s" % os.path.basename(temperaturefile))
			self.temperature_dss += [DataSet(temperaturefile, do_smooth=sg_enabled, smooth_kernel=sg_kernel, smooth_order=sg_order)]

		# TODO: correlation

		status_cb("Averaging strain data")
		self.strain_avg_ds = AverageDataSet(self.strain_dss, name="Average strain")

		status_cb("Averaging temperature data")
		self.temperature_avg_ds = AverageDataSet(self.temperature_dss, name="Average temperature")

		status_cb("Applying temperature compensation")
		self.compensated_ds = TemperatureCompensatedDataSet(strain=self.strain_avg_ds, temperature=self.temperature_avg_ds, temperature_factor=temperature_factor, name="Compensated strain")

		finished_cb()

		graph = StandaloneGraph(self.compensated_ds, extradss=[self.strain_avg_ds, self.temperature_avg_ds])
		graph.show()


if __name__ == "__main__":
	ui = StrainAnalyserMainUI()
	ui.show()
	gtk.main()

