#!/usr/bin/python

DO_CORRELATE = True

import sys
import numpy

import matplotlib
matplotlib.use('Qt4Agg')

import matplotlib.pyplot as pyplot

from savitzkygolay import savitzky_golay

class DataSet:
	do_smooth = True
	smooth_order = 5
	smooth_kernel = 51

	class FileFormatException(Exception):
		pass

	def __init__(self, filename):
		self.properties = {}
		self.name = filename

		with open(filename, 'r') as f:
			if f.next().rstrip() != "Strain":
				raise self.__class__.FileFormatException()

			for line in f:
				line = line.strip()
				if line == 'DATA':
					break
				else:
					key,val = line.split(',')
					self.properties[key] = val

			self.step = float(self.properties['RESOLUTION[m]'])
			self.len = int(self.properties['DATA NUM'])
			self.y = numpy.fromiter(f, numpy.float, self.len)
			self.diff = numpy.fabs(numpy.diff(self.y))

			self.locate_end()
			self.smooth()

	def locate_end(self):
		# TODO: this heuristic is just a quick hack and doesn't work well
		threshold = numpy.max(self.diff) * 0.4
		self.end = numpy.argwhere(self.diff > threshold)[0] - 1

	def smooth(self):
		if self.do_smooth:
			if y_orig not in self:
				self.y_orig = self.y
		self.y = savitzky_golay(self.y_orig, 51, 5)


if __name__ == "__main__":
	dss = []
	for filename in sys.argv[1:]:
		dss += [DataSet(filename)]
	fig = pyplot.figure()
	axes = fig.add_subplot(1,1,1)
	axes.grid(True)
	axes.axvline(x=0, color='black')
	axes.axhline(y=0, color='black')
	axes.set_xlabel("Distance (m)")
	axes.set_ylabel("Strain (%)")
	xmax = 1000000
	ymin = 0
	ymax = 0
	plotend = 1000000
	ds_correlate = None
	for ds in dss:
		if DO_CORRELATE and ds_correlate != None:
			# datasets may be of different resolutions, and of different lengths!
			# correlation seems to work poorly with different lengths
			# trim the ends to try to keep noise out
			correlation_step = max(ds_correlate.step, ds.step)
			correlation_length = min(ds.end * ds.step, ds_correlate.end * ds_correlate.step)
			correlation_data1 = numpy.interp(numpy.arange(correlation_length*0.2, correlation_length*0.8, correlation_step), ds_correlate.step * numpy.arange(0, ds_correlate.end), ds_correlate.y[:ds_correlate.end])
			correlation_data2 = numpy.interp(numpy.arange(correlation_length*0.2, correlation_length*0.8, correlation_step), ds.step * numpy.arange(0, ds.end), ds.y[:ds.end])
			correlation = numpy.correlate(correlation_data1, correlation_data2, mode='full')
			shift = correlation_step * (numpy.argmax(correlation) - len(correlation)/2)
		else:
			shift = 0
			ds_correlate = ds
		axes.plot([i*ds.step + shift for i in range(ds.len)], ds.y, label=ds.name)
		xmax = min(xmax, ds.step*ds.end + shift)
		plotend = min(plotend, ds.end + int(shift/ds.step))
	for ds in dss:
		ymax = max(ymax, ds.y[:plotend].max())
		ymin = min(ymin, ds.y[:plotend].min())
	axes.axvline(xmax, color='r')
	axes.set_xlim(0, xmax)
	axes.set_ylim(ymin - (ymax-ymin)*0.1, ymax + (ymax-ymin)*0.1)

	leg = axes.legend(labelspacing=0)
	for t in leg.get_texts():
		t.set_fontsize('small')

	pyplot.show()
