#!/usr/bin/python

import os.path

import pygtk
pygtk.require('2.0')
import gtk
import gtk.glade

import matplotlib
matplotlib.use('GTKAgg')
import matplotlib.pyplot as pyplot
from matplotlib.backends.backend_gtkagg import NavigationToolbar2GTKAgg as NavigationToolbar
from matplotlib.backends.backend_gtkagg import FigureCanvasGTKAgg as FigureCanvas

import numpy

from savitzkygolay import savitzky_golay

class DataSet:
	class FileFormatException(Exception):
		pass

	def __init__(self, filename, do_smooth, smooth_kernel, smooth_order):
		self.properties = {}
		self.name, _ = os.path.splitext(os.path.basename(filename))

		with open(filename, 'r') as f:
			if f.next().rstrip() != "Strain":
				raise self.__class__.FileFormatException()

			for line in f:
				line = line.strip()
				if line == 'DATA':
					break
				else:
					key,val = line.split(',')
					self.properties[key] = val

			self.step = float(self.properties['RESOLUTION[m]'])
			self.len = int(self.properties['DATA NUM'])
			self.y = numpy.fromiter(f, numpy.float, self.len)
			self.diff = numpy.fabs(numpy.diff(self.y))

			self.locate_end()
			self.smooth(enabled=do_smooth, kernel=smooth_kernel, order=smooth_order)

	def locate_end(self):
		# TODO: this heuristic is just a quick hack and doesn't work well
		threshold = numpy.max(self.diff) * 0.4
		self.end = numpy.argwhere(self.diff > threshold)[0] - 1

	def smooth(self, order, kernel, enabled):
		if enabled:
			if not hasattr(self, 'y_orig'):
				self.y_orig = self.y
			self.y = savitzky_golay(self.y_orig, kernel, order)
		else:
			if hasattr(self, 'y_orig'):
				self.y = self.y_orig


class MainGui:
	def __init__(self):
		self.wTree = gtk.glade.XML("strainanalyser.glade")
		handlers = {
			"on_noisefilt_toggled": self.toggle_strain_noisefilt,
			"on_noisefilt_order_value_changed": self.update_strain_noisefilt_order,
			"on_noisefilt_kernel_value_changed": self.update_strain_noisefilt_kernel,
			"on_filechooser_selection_changed": self.update_strain_graph,
			"on_main_destroy": gtk.main_quit
		}
		self.wTree.signal_autoconnect(handlers)

		self.win = self.wTree.get_widget("main")

		self.graphcontainer = self.wTree.get_widget("graphcontainer")
		self.datagraph = DataGraph(self.win, self.graphcontainer)

	def toggle_strain_noisefilt(self, widget):
		print "toggle_strain_noisefilt: %s" % widget
		self.datagraph.update_smoothing(enabled=widget.get_active())

	def update_strain_noisefilt_order(self, widget):
		print "update_strain_noisefilt_order: %s" % widget
		val = widget.get_value()
		kernelctrl = self.wTree.get_widget("noisefilt_kernel")
		kernelctrl.set_range(val*2 + 1, 99)
		self.datagraph.update_smoothing(order=val)

	def update_strain_noisefilt_kernel(self, widget):
		print "update_strain_noisefilt_kernel: %s" % widget
		self.datagraph.update_smoothing(kernel=widget.get_value())

	def update_strain_graph(self, widget):
		print "update_strain_graph: %s" % widget
		self.datagraph.setfiles(widget.get_filenames())


class DataGraph:
	do_smooth = True
	smooth_order = 5
	smooth_kernel = 51

	def __init__(self, win, vbox):
		self.graphcontainer = vbox

		self.fig = pyplot.figure()
		self.axes = self.fig.add_subplot(1,1,1)
		self.canvas = FigureCanvas(self.fig) # a gtk.DrawingArea
		self.canvas.show()

		self.toolbar = NavigationToolbar(self.canvas, win)

		self.graphcontainer.pack_start(self.toolbar, False, False)
		self.graphcontainer.pack_start(self.canvas)

		self.dss = []

	def update_smoothing(self, order=None, kernel=None, enabled=None):
		if order != None:
			self.smooth_order = order
		if kernel != None:
			self.smooth_kernel = kernel
		if enabled != None:
			self.do_smooth = enabled

		for ds in self.dss:
			ds.smooth(order=self.smooth_order, kernel=self.smooth_kernel, enabled=self.do_smooth)
		self.redraw()

	def setfiles(self, filenames):
		self.dss = []
		for filename in filenames:
			try:
				self.dss += [DataSet(filename=filename, do_smooth=self.do_smooth, smooth_order=self.smooth_order, smooth_kernel=self.smooth_kernel)]
			except Exception, e:
				print "Not drawing %s: %s" % (filename, e)

		self.redraw()

	def redraw(self):
		DO_CORRELATE = True

		self.axes.clear()
		self.axes.grid(True)
		self.axes.axvline(x=0, color='black')
		self.axes.axhline(y=0, color='black')
		self.axes.set_xlabel("Distance (m)")
		self.axes.set_ylabel("Strain (%)")

		xmax = 1000000
		ymin = 0
		ymax = 0
		plotend = 1000000
		ds_correlate = None
		for ds in self.dss:
			if DO_CORRELATE and ds_correlate != None:
				# datasets may be of different resolutions, and of different lengths!
				# correlation seems to work poorly with different lengths
				# trim the ends to try to keep noise out
				correlation_step = max(ds_correlate.step, ds.step)
				correlation_length = min(ds.end * ds.step, ds_correlate.end * ds_correlate.step)
				correlation_data1 = numpy.interp(numpy.arange(correlation_length*0.2, correlation_length*0.8, correlation_step), ds_correlate.step * numpy.arange(0, ds_correlate.end), ds_correlate.y[:ds_correlate.end])
				correlation_data2 = numpy.interp(numpy.arange(correlation_length*0.2, correlation_length*0.8, correlation_step), ds.step * numpy.arange(0, ds.end), ds.y[:ds.end])
				correlation = numpy.correlate(correlation_data1, correlation_data2, mode='full')
				shift = correlation_step * (numpy.argmax(correlation) - len(correlation)/2)
			else:
				shift = 0
				ds_correlate = ds
			self.axes.plot([i*ds.step + shift for i in range(ds.len)], ds.y, label=ds.name)
			xmax = min(xmax, ds.step*ds.end + shift)
			plotend = min(plotend, ds.end + int(shift/ds.step))
		for ds in self.dss:
			ymax = max(ymax, ds.y[:plotend].max())
			ymin = min(ymin, ds.y[:plotend].min())
		self.axes.axvline(xmax, color='r')
		self.axes.set_xlim(0, xmax)
		self.axes.set_ylim(ymin - (ymax-ymin)*0.1, ymax + (ymax-ymin)*0.1)

		if len(self.dss) > 0:
			leg = self.axes.legend(labelspacing=0)
			for t in leg.get_texts():
				t.set_fontsize('small')

		self.canvas.draw()


app = MainGui()
gtk.main()
