#!/usr/bin/python

import os
import numpy
import gtk
import gobject


class StrainAnalyserDatafile(gobject.GObject):

	class FileFormatException(Exception):
		def __str__(self):
			return "Unrecognised file format"

	def __init__(self, strain_filename, loss_filename):
		self.properties = {}
		self.strain_name, _ = os.path.splitext(os.path.basename(strain_filename))
		#self.loss_name, _ = os.path.splitext(os.path.basename(loss_filename))
		self.loss_name = "synthesised loss data"

		with open(strain_filename, 'r') as f:
			if f.next().rstrip() != "Strain":
				raise self.__class__.FileFormatException(strain_filename)

			for line in f:
				line = line.strip()
				if line == 'DATA':
					break
				else:
					key,val = line.split(',')
					self.properties[key] = val

			self.resolution = float(self.properties['RESOLUTION[m]'])
			self.refractiveindex = float(self.properties['IOR'])
			self.len = int(self.properties['DATA NUM'])
			self.straindata = numpy.fromiter(f, numpy.float, self.len) * 10000 # percent -> microstrain
			assert(len(self.straindata) == self.len)

		# TODO: something with the loss data!
		# For now, synthesise some
		diff = numpy.fabs(numpy.diff(self.straindata))
		threshold = numpy.max(diff) * 0.4
		end = numpy.argwhere(diff > threshold)[0][0] - 1
		self.confidence = numpy.array([1] * end + [0] * (self.len - end))

		mask = numpy.ma.masked_less(self.confidence, 0.1).mask
		self.straindata = numpy.ma.MaskedArray(self.straindata, mask=mask)

	def __len__(self):
		return self.len


class StrainAnalyserDataset(gtk.ListStore):

	def __init__(self):
		super(StrainAnalyserDataset, self).__init__(str, str, gobject.TYPE_PYOBJECT, gobject.TYPE_BOOLEAN)

	def add_file(self, data):
		self.append([data.strain_name, data.loss_name, data, True])

	def remove_file(self, index):
		self.remove(self.iter_nth_child(None, index))


class StrainAnalyserAveragedDataset(object):

	def __init__(self, dataset):
		self.dataset = dataset
		self.len = 0
		self.resolution_factor = 1.

	def refresh(self):
		# correlation
		firstrun = True
		for _,_,datafile,enabled in self.dataset:
			if not enabled:
				continue
			if firstrun:
				firstrun = False
				ds_correlate = datafile.straindata
				self.properties = datafile.properties
				self.refractiveindex = datafile.refractiveindex
				self.resolution = datafile.resolution
				self.len = datafile.len
				# this will become a 2D matrix of the data and weights to average
				data = numpy.ma.MaskedArray([datafile.straindata])
				weight = numpy.array([datafile.confidence])
			else:
				assert(self.resolution == datafile.resolution)
				assert(self.len == datafile.len)
				correlation = numpy.correlate(
						numpy.ma.MaskedArray(ds_correlate, mask=datafile.straindata.mask).compressed(),
						numpy.ma.MaskedArray(datafile.straindata, mask=ds_correlate.mask).compressed(),
						mode='full'
					)
				shift = numpy.argmax(correlation) - len(correlation)/2
				if shift >= 0:
					# shunt the data along by adding extra masked data points (and then trim the far end)
					ds = numpy.ma.concatenate([numpy.ma.masked_all(shift, int), datafile.straindata])[:self.len]
					confidence = numpy.ma.concatenate([[0]*shift, datafile.confidence])[:self.len]
				else:
					# chop some data points off the near end (and then add some blank points at the far end)
					ds = numpy.ma.concatenate([datafile.straindata[-shift:], numpy.ma.masked_all(-shift, int)])
					confidence = numpy.ma.concatenate([datafile.confidence[-shift:], [0]*(-shift)])

				# build up those 2D matrices
				data = numpy.vstack((data, [ds]))
				weight = numpy.vstack((weight, [confidence]))

		if not firstrun: # did we actually include any data?
			# averaging
			self.straindata, self.confidence = numpy.ma.average(axis=0, a=data, weights=weight, returned=True)
			assert(self.len == len(self.straindata))
		else:
			self.len = 0

	def __len__(self):
		return self.len


class StrainAnalyserTemperatureCompensatedDataset(object):

	def __init__(self, strain_dataset, temperature_dataset):
		self.strain_dataset = strain_dataset
		self.temperature_dataset = temperature_dataset

		self.name = "Untitled"

		# Constants in units of 10^-6 / degree C
		self.alpha_a = 19.47      # const, from Mohamad's thesis
		self.alpha_n_temp = 4.2   # Unitube cable, from Mohamad's thesis
		self.alpha_n_strain = 10  # estimate of reasonable alpha_concrete, EN1992-1-1 / Mohamad

		self.temperature_offset = 0
		self.range_start = None
		self.range_end = None

		self.refractiveindex = None

		self.len = 0

	def update_resolution(self):
		self.strain_dataset.refresh()
		self.temperature_dataset.refresh()

		if len(self.strain_dataset) > 0:
			template = self.strain_dataset
		else:
			template = self.temperature_dataset

		self.properties = template.properties
		self.recorded_resolution = template.resolution
		self.recorded_refractiveindex = template.refractiveindex
		if self.refractiveindex == None:
			self.refractiveindex = template.refractiveindex
		self.resolution = template.resolution * self.refractiveindex / self.recorded_refractiveindex

	def refresh(self):
		assert(self.strain_dataset.resolution == self.temperature_dataset.resolution)
		assert(self.strain_dataset.refractiveindex == self.temperature_dataset.refractiveindex)
		assert(self.strain_dataset.len == self.temperature_dataset.len)

		self.update_resolution()

		self.len = self.strain_dataset.len

		# Deal with temperature offset
		shift = self.temperature_offset / self.temperature_dataset.resolution
		if shift >= 0:
			# shunt the data along by adding extra masked data points
			temperature_data_aligned = numpy.ma.concatenate(
					[numpy.ma.masked_all(shift, int),
					self.temperature_dataset.straindata]
				)[:self.temperature_dataset.len]
		else:
			# chop some data points off the near end
			temperature_data_aligned = self.temperature_dataset.straindata[shift:]

		assert(len(temperature_data_aligned) == len(self.strain_dataset.straindata))

		# Truncate to range of interest
		if self.range_start != None:
			range_start_idx = int(self.range_start / self.resolution)
			range_end_idx = int(self.range_end / self.resolution + 1)
			temperature_data_aligned = temperature_data_aligned[range_start_idx:range_end_idx]
			strain_data_aligned = self.strain_dataset.straindata[range_start_idx:range_end_idx]
			self.len = range_end_idx - range_start_idx
		else:
			strain_data_aligned = self.strain_dataset.straindata

		# Obtain mechanical strain from mechanical+thermal strain and thermal strain datasets.
		# The following equation is the whole point of this project. :-)
		#
		# NB: numpy arrays multiply each element individually -- cf normal python arrays which do not!
		# These should be numpy arrays (in fact numpy.ma.MaskedArrays).
		self.straindata = (strain_data_aligned -
				temperature_data_aligned * (self.alpha_a + self.alpha_n_strain)/(self.alpha_a + self.alpha_n_temp))

	def __len__(self):
		return self.len


class StrainAnalyserDatasetCollection(gtk.ListStore):

	def __init__(self):
		super(StrainAnalyserDatasetCollection, self).__init__(str, gobject.TYPE_BOOLEAN, gobject.TYPE_PYOBJECT, str)

	def add(self, data):
		self.append([data.name, False, data, data.resolution])

	def remove(self, index):
		super(StrainAnalyserDatasetCollection, self).remove(self.iter_nth_child(None, index))

	def move_up(self, index):
		if index > 0:
			self.swap(
					self.iter_nth_child(None, index),
					self.iter_nth_child(None, index-1)
				)

	def move_down(self, index):
		if index < (len(self) - 1):
			self.swap(
					self.iter_nth_child(None, index),
					self.iter_nth_child(None, index+1)
				)

	def get_reference(self):
		"Return the StrainAnalyserDataset which is currently marked as the reference, or None."
		for row in self:
			if row[1]: # if reference radio button is active
				return row[2] # return the dataset payload
		return None


class StrainAnalyserDiffDataset(object):

	def __init__(self, original_ds, diff_array, length, resolution):
		self.resolution = resolution
		self.refractiveindex = original_ds.refractiveindex
		self.properties = original_ds.properties
		self.name = original_ds.name

		self.straindata = diff_array
		self.len = length
		assert(self.len == len(diff_array))

	def __len__(self):
		return self.len


class StrainAnalyserDiffDatasetCollection(object):

	def __init__(self, dataset_collection):
		self.dataset_collection = dataset_collection

		self.noisefilt_enabled = False
		self.noisefilt_order = 5
		self.noisefilt_kernel = 51

		self.diffed_data = []

	def refresh(self):
		if len(self.dataset_collection) == 0:
			return

		ref = self.dataset_collection.get_reference()
		if ref is not None:
			correlation_base = ref
		else:
			correlation_base = self.dataset_collection[0][2]

		self.resolution = correlation_base.resolution

		self.diffed_data = []

		for _,_,ds,enabled in self.dataset_collection:

			if not enabled:
				continue

			# resample to match the reference
			y1 = numpy.arange(0, ds.len*ds.resolution, correlation_base.resolution)
			y2 = numpy.arange(0, ds.len*ds.resolution, ds.resolution)
			# may lose (or attempt to gain!) a point due to rounding; never mind
			y1 = y1[:min(len(y1), len(y2), len(ds.straindata))]
			y2 = y2[:min(len(y1), len(y2), len(ds.straindata))]
			straindata = ds.straindata[:min(len(y1), len(y2), len(ds.straindata))]
			resampled_data = numpy.interp(y1, y2, straindata)

			# trim the ends for correlation
			# and force into a MaskedArray in case it's not already
			intersection = min(len(correlation_base.straindata), len(resampled_data))
			correlation_data_ref = numpy.ma.MaskedArray(correlation_base.straindata[0.2*intersection:0.8*intersection])
			correlation_data = numpy.ma.MaskedArray(resampled_data[0.2*intersection:0.8*intersection])
			assert(len(correlation_data_ref) == len(correlation_data))

			# correlate
			correlation = numpy.correlate(
					numpy.ma.MaskedArray(correlation_data_ref, mask=correlation_data.mask).compressed(),
					numpy.ma.MaskedArray(correlation_data, mask=correlation_data_ref.mask).compressed(),
					mode='full'
				)

			# shift
			# TODO: this is better than the shift code above; the code above may need replacing with this
			shift = numpy.argmax(correlation) - len(correlation)/2
			if shift >= 0:
				# shunt the data along by adding extra masked data points (and then trim the far end)
				new_straindata = numpy.ma.concatenate([numpy.ma.masked_all(shift, int), ds.straindata])[:intersection]
			else:
				# chop some data points off the near end (and pad at the far end)
				new_straindata = numpy.ma.concatenate([ds.straindata, numpy.ma.masked_all(-shift, int)])[:intersection]

			# diff
			if ref is not None:
				new_straindata_diff = new_straindata - ref.straindata
			else:
				new_straindata_diff = new_straindata

			# TODO: smoothing

			self.diffed_data += [StrainAnalyserDiffDataset(
					original_ds=ds,
					length=intersection,
					diff_array=new_straindata_diff,
					resolution=self.resolution
				)]

