#!/usr/bin/python
# coding: utf-8

import numpy
import gtk

import matplotlib
import matplotlib.pyplot as pyplot
from matplotlib.backends.backend_gtkcairo import FigureCanvasGTKCairo as FigureCanvas
from matplotlib.backends.backend_gtk import NavigationToolbar2GTK as NavigationToolbar


class StrainAnalyserDatasetDiffGraph:

	def __init__(self, container_vbox, dataset_diff_collection):
		self.fig = pyplot.figure()
		self.axes = self.fig.add_subplot(1,1,1, autoscale_on=False)
		self.canvas = FigureCanvas(self.fig) # a gtk.DrawingArea
		self.canvas.show()

		container_vbox.pack_start(self.canvas)
		toolbar = NavigationToolbar(self.canvas, container_vbox.get_parent_window())
		container_vbox.pack_start(toolbar, False, False)

		self.dataset_diff_collection = dataset_diff_collection

		self.redraw()

	def redraw(self):
		self.dataset_diff_collection.refresh()

		self.axes.clear()
		self.axes.grid(True)
		self.axes.axvline(x=0, color='black')
		self.axes.axhline(y=0, color='black')
		self.axes.set_xlabel("Distance (m)")
		self.axes.set_ylabel("Strain (µε)")

		empty = True
		ymax = 0
		ymin = 0
		for dataset in self.dataset_diff_collection.diffed_data:
			self.axes.plot(
					[i*dataset.resolution for i in range(dataset.len)],
					dataset.straindata,
					label=dataset.name
				)
			empty = False
			ymax = max(ymax, dataset.straindata.max())
			ymin = min(ymin, dataset.straindata.min())

		# TODO: deal with selected region

		if not empty:
			self.axes.set_ylim(ymin - (ymax-ymin)*0.1, ymax + (ymax-ymin)*0.1)
			#if self.straindataset.len > 0:
			#	self.axes.set_xlim(0, numpy.ma.notmasked_edges(self.straindataset.straindata)[1] * self.straindataset.resolution)

			leg = self.axes.legend(labelspacing=0)

		self.canvas.draw()


class StrainAnalyserDataGraph:

	def __init__(self, widget_add_cb, set_range_cb, set_temperature_offset_cb,
			avg_dataset_strain, avg_dataset_temperature):
		self.set_range_cb = set_range_cb
		self.set_temperature_offset_cb = set_temperature_offset_cb

		self.fig = pyplot.figure()
		self.axes = self.fig.add_subplot(1,1,1, autoscale_on=False)
		self.canvas = FigureCanvas(self.fig) # a gtk.DrawingArea
		self.canvas.show()

		widget_add_cb(self.canvas)

		self.straindataset = avg_dataset_strain
		self.temperaturedataset = avg_dataset_temperature

		self.range_mode = False
		self.press = None
		self.temperature_shift = 0
		self.range_start = None
		self.range_end = None

		self.event_cid = {}
		self.event_cid['button_press_event'] = self.canvas.mpl_connect('button_press_event', self.on_press)
		self.event_cid['button_release_event'] = self.canvas.mpl_connect('button_release_event', self.on_release)
		self.event_cid['motion_notify_event'] = self.canvas.mpl_connect('motion_notify_event', self.on_motion)

		self.redraw()

	def redraw(self):
		self.straindataset.refresh()
		self.temperaturedataset.refresh()

		self.axes.clear()
		self.axes.grid(True)
		self.axes.axvline(x=0, color='black')
		self.axes.axhline(y=0, color='black')
		self.axes.set_xlabel("Distance (m)")
		self.axes.set_ylabel("Strain (µε)")

		empty = True
		ymax = 0
		ymin = 0
		if self.straindataset.len > 0:
			self.axes.plot(
					[i*self.straindataset.resolution*self.straindataset.resolution_factor
							for i in range(self.straindataset.len)],
					self.straindataset.straindata,
					label="Combined strain",
					color="blue"
				)
			empty = False
			ymax = max(ymax, self.straindataset.straindata.max())
			ymin = min(ymin, self.straindataset.straindata.min())
		if self.temperaturedataset.len > 0:
			self.axes.plot(
					[i*self.temperaturedataset.resolution*self.temperaturedataset.resolution_factor + self.temperature_shift
							for i in range(self.temperaturedataset.len)],
					self.temperaturedataset.straindata,
					label="Thermal strain",
					color="red"
				)
			empty = False
			ymax = max(ymax, self.temperaturedataset.straindata.max())
			ymin = min(ymin, self.temperaturedataset.straindata.min())

		if not empty:
			if self.range_start != None and self.range_end != None:
				self.axes.axvspan(self.range_start, self.range_end, color='green', alpha=0.25)

			self.axes.set_ylim(ymin - (ymax-ymin)*0.1, ymax + (ymax-ymin)*0.1)
			if self.straindataset.len > 0:
				self.axes.set_xlim(0, numpy.ma.notmasked_edges(self.straindataset.straindata)[1] * self.straindataset.resolution)

			leg = self.axes.legend(labelspacing=0)

		self.canvas.draw()

	def set_range_mode(self):
		self.range_mode = True

	def set_align_mode(self):
		self.range_mode = False

	def set_temperature_shift(self, shift):
		self.temperature_shift = shift
		self.redraw()
		self.set_temperature_offset_cb(shift)

	def set_range(self, x1, x2):
		if x1 < x2:
			self.range_start = x1
			self.range_end = x2
		else:
			self.range_start = x2
			self.range_end = x1
		self.redraw()
		self.set_range_cb(self.range_start, self.range_end)

	# draggability:
	# http://matplotlib.sourceforge.net/users/event_handling.html

	def on_press(self, event):
		if event.inaxes != self.axes: return
		if self.range_mode:
			if self.straindataset.len == 0 and self.temperaturedataset.len == 0: return
		else:
			if self.straindataset.len == 0 or self.temperaturedataset.len == 0: return
		self.press = self.temperature_shift, event.xdata, event.ydata

		if self.range_mode:
			self.set_range(event.xdata, event.xdata)

	def on_motion(self, event):
		if self.press is None or event.inaxes != self.axes:
			return
		orig_shift, xpress, ypress = self.press

		if self.range_mode:
			self.set_range(event.xdata, xpress)
		else:
			dx = event.xdata - xpress
			self.set_temperature_shift(orig_shift + dx)

	def on_release(self, event):
		self.press = None


def simple_graph_viewer(ds):
	win = gtk.Window()
	win.set_default_size(1000,700)
	win.set_title("%s - Graph" % ds.name)
	vbox = gtk.VBox()
	win.add(vbox)
	win.show_all()

	fig = pyplot.figure()
	axes = fig.add_subplot(1,1,1)
	canvas = FigureCanvas(fig)
	canvas.show()

	vbox.pack_start(canvas)
	toolbar = NavigationToolbar(canvas, win)
	vbox.pack_start(toolbar, False, False)

	axes.grid(True)
	axes.axvline(x=0, color='black')
	axes.axhline(y=0, color='black')
	axes.set_xlabel("Distance (m)")
	axes.set_ylabel("Strain (µε)")
	axes.plot([i*ds.resolution for i in range(ds.len)], ds.straindata, label=ds.name)
	axes.set_ylim(ds.straindata.min() - 50, ds.straindata.max() + 50)

	canvas.draw()

